/************************************************************************
 *   Quake Internet Relay Chat Daemon, src/quake.c
 *   Copyright (C) 1990 Jarkko Oikarinen and
 *                      University of Oulu, Computing Center
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 1, or (at your option)
 *   any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *   $Id: quake.c,v 1.99 2004/07/11 07:06:09 tirc Exp $
 */

#include "config.h"
#include "struct.h"
#include "common.h"
#include "sys.h"
#include "numeric.h"
#include "msg.h"
#include "channel.h"
#include <time.h>
#include <sys/stat.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#ifdef _WIN32
#include <io.h>
#endif
#include <fcntl.h>
#include "h.h"
#ifdef STRIPBADWORDS
#include "badwords.h"
#endif
#ifdef _WIN32
#include "version.h"
#endif

#ifdef EAS
void eas_start();
void eas_stop();

EVENT(eas_check);
Event *eas_event;

void eas_inc();
void eas_clear();
int eas_state();
void eas_set();

int eas_count = 0;
int eas_level = 0;
int eas_timer;

/*
 * eas_inc
 * Incremements the eas counter
 */

void eas_inc()
{
	if (EventFind("eas"))
	eas_count++;
}

/*
 * eas_clear
 * Resets the eas counter
 */

void eas_clear()
{
	eas_count = 0;
}

/*
 * eas_set
 * Sets and Removes Emergency Alert System abilities
 */

void eas_set()
{
	if ((eas_level == 1) || (eas_level == 2))
	{
		eas_timer--;
	}
	if (eas_timer == 0)
	{
			if (eas_level == 1)
				sendto_realops("%s - Emergency Alert System - Botnet Watch Expired", me.name);
			if (eas_level == 2)
				sendto_realops("%s - Emergency Alert System - Botnet Warning Expired", me.name);
                        eas_level = 0;
			eas_timer = EASTIME;
	}
	else
	{
        if (eas_count > EASLOCK)
        {
                eas_level = 2;
                sendto_realops("%s - Emergency Alert System - Botnet Warning - %i Connections", me.name, eas_count);
                sendto_realops("%s - Emergency Alert System - Lockdown for %i seconds", me.name, (eas_timer * EASLOOP));
        }
        else if (eas_count > EASWARN)
        {
                eas_level = 1;
                sendto_realops("%s - Emergency Alert System - Botnet Watch - %i Connections", me.name, eas_count);
        }
	}
}

/*
 * eas_state
 * Returns the current state
 * Normal, Warning or Watch
 */

int eas_state()
{
	if (eas_level == 1)
		return 1;
	if (eas_level == 2)
		return 2;
	return 0;
}

/*
 * eas_control
 * eas control function
 * on/off
 */

CMD_FUNC(eas_control)
{
	if (!(IsAdmin(sptr) || IsCoAdmin(sptr))) {
                sendto_one(sptr, err_str(ERR_NOPRIVILEGES), me.name, parv[0]);
                return -1;
        }

	if (!parv[1])
	{
		sendto_realops("%s - Emergency Alert System Control - Need Option", me.name);
		return 0;
	}

	if (!stricmp(parv[1], "ON"))
	{
		eas_start();
	}
	else if (!stricmp(parv[1], "OFF"))
	{
		eas_stop();
	}
	else {
		sendto_realops("%s - Emergency Alert System Control - Need Option", me.name);
	}
	return 0;
}

/*
 * eas_start
 * enables eas
 */

void eas_start()
{
	if (EventFind("eas"))
	sendto_realops("%s - Emergency Alert System is Already Activated", me.name);
	else {
	eas_timer = EASTIME;
	sendto_realops("%s - Emergency Alert System Activated", me.name);
	LockEventSystem();
	eas_event = EventAddEx(NULL, "eas", EASLOOP, 0, eas_check, NULL);
	UnlockEventSystem();
	}
}

/*
 * eas_stop
 * disables eas
 */

void eas_stop()
{
	if (!EventFind("eas"))
	sendto_realops("%s - Emergency Alert System is Already Disabled", me.name);
	else {
	sendto_realops("%s - Emergency Alert System - All Active Watches and Warnings Expired", me.name);
	sendto_realops("%s - Emergency Alert System Disabled", me.name);
	LockEventSystem();	
	EventDel(eas_event);
	UnlockEventSystem();
	eas_level = 0;
	eas_count = 0;
	eas_timer = EASTIME;
	}
}

/*
 * eas_check
 * eas loop event
 */

EVENT(eas_check)
{
	eas_set();
	eas_clear();
}

#endif

#ifdef CENTRAL_CONF
void m_send_conf(aClient *cptr);

/*
** m_get_conf
** 	Asking for the remote confs
**
*/

CMD_FUNC(m_get_conf)
{
	aClient *cconfbuf;
	int fail = 0;

	if (!(IsAdmin(sptr) || IsCoAdmin(sptr))) {
                sendto_one(sptr, err_str(ERR_NOPRIVILEGES), me.name, parv[0]);
                return -1;
        }

	sendto_snomask(SNO_CONF, " %s - CentralConf - Requesting Remote Configuration", me.name);

	cconfbuf = find_server(MASTERSERVER, NULL);

	if (cconfbuf)
	{
#ifdef CENTRALCLEAN
	(void)m_central_clean(cconfbuf);
#else
	sendto_one(cconfbuf, ":%s REQCONF %s :%s", me.name, cconfbuf->name, me.name);
#endif
	sendto_snomask(SNO_CONF, " %s - CentralConf - Remote Configuration Request Sent to %s", me.name, cconfbuf->name);
	return 0;
	}
	else {
		sendto_snomask(SNO_CONF, " %s - CentralConf - Master Server unreachable, trying backup", me.name);
	}
	
	cconfbuf = find_server(BACKUPSERVER, NULL);
	if (cconfbuf)
	{
#ifdef CENTRALCLEAN
        (void)m_central_clean(cconfbuf);
#else
        sendto_one(cconfbuf, ":%s REQCONF %s :%s", me.name, cconfbuf->name, me.name);
#endif
	sendto_snomask(SNO_CONF, " %s - CentralConf - Remote Configuration Request Sent to %s", me.name, cconfbuf->name);
        return 0;
        }
	if (parv[1])
	{
		cconfbuf = find_server(parv[1], NULL);
		if (cconfbuf)
		{
		sendto_snomask(SNO_CONF, " %s - CentralConf - Backup Server unreachable, trying %s", me.name, parv[1]);
#ifdef CENTRALCLEAN
		(void)m_central_clean(cconfbuf);
#else
		sendto_one(cconfbuf, ":%s REQCONF %s :%s", me.name, cconfbuf->name, me.name);	
#endif
		sendto_snomask(SNO_CONF, " %s - CentralConf - Remote Configuration Request Sent to %s", me.name, cconfbuf->name);
		return 0;
		}
	}
	sendto_snomask(SNO_CONF, " %s - CentralConf - Servers unreachable, specify a valid server", me.name);
	return 0;
}
/*
** m_conf_req
** 	Recieved a request for remote confs
*/

CMD_FUNC(m_conf_req)
{
        aClient *cconfbuf;

	if (!parv[1])
        {
                sendto_snomask(SNO_CONF, " %s - CentralConf - Dropped Malformed Configuration Request", me.name);
                return 0;
        }

        if(stricmp(parv[1], me.name))
        {
                cconfbuf = find_server(parv[1], NULL);
                if (cconfbuf)
                        sendto_one(cconfbuf, ":%s REQCONF %s :%s", sptr->name, cconfbuf->name, sptr->name);
                return 0;
        }

	if (IsServer(sptr))
	{
		sendto_snomask(SNO_CONF, " %s - CentralConf - Remote Configuration Request Recieved from %s", me.name, sptr->name);
		m_send_conf(sptr);
	}
	return 0;
}

#ifdef CENTRALCLEAN
/*
** m_central_clean
**	rehash before asking for new confs
*/

void m_central_clean(aClient *cptr)
{
	(void)rehash(&me, &me, 2);	
	sendto_one(cptr, ":%s REQCONF %s :%s", me.name, cptr->name, me.name);
}
#endif

/*
** m_send_conf
**      multiple buffers for what to send
*/

void m_send_conf(aClient *cptr)
{
	FILE *cstream;
	char buf[1024];
	char *nl;

	cstream = fopen(CENTRALFILE, "rt");

	if (cstream)
	{
		buf[1024] = 0;

		sendto_snomask(SNO_CONF, " %s - CentralConf - Sending Remote Configuration to %s", me.name, cptr->name);

		while (fgets(buf, 1024, cstream))
		{
			nl = strchr(buf, '\n');
			if (nl)
				*nl = 0;
			if (!(buf[0] == '#'))
			sendto_one(cptr, ":%s READCONF %s :%s", me.name, cptr->name, buf);
		}
		fclose(cstream);
	}
	else
	{
		sendto_snomask(SNO_CONF, " %s - CentralConf - Error Sending Remote Configuration to %s", me.name, cptr->name);
	}
}

/*
** m_read_conf
**      parv[0] = sender prefix
**      parv[1] = me
**	parv[2] = colon seperated string
*/

#define ircstrdup(x,y) if (x) MyFree(x); if (!y) x = NULL; else x = strdup(y)
#define ircfree(x) if (x) MyFree(x); x = NULL

static int _OldOperFlags[] = {
        OFLAG_LOCAL, 'o',
        OFLAG_GLOBAL, 'O',
        OFLAG_REHASH, 'r',
        OFLAG_DIE, 'D',
        OFLAG_RESTART, 'R',
        OFLAG_HELPOP, 'h',
        OFLAG_GLOBOP, 'g',
        OFLAG_WALLOP, 'w',
        OFLAG_LOCOP, 'l',
        OFLAG_LROUTE, 'c',
        OFLAG_GROUTE, 'L',
        OFLAG_LKILL, 'k',
        OFLAG_GKILL, 'K',
        OFLAG_KLINE, 'b',
        OFLAG_UNKLINE, 'B',
        OFLAG_LNOTICE, 'n',
        OFLAG_GNOTICE, 'G',
        OFLAG_ADMIN_, 'A',
        OFLAG_SADMIN_, 'a',
	OFLAG_NADMIN, 'N',
        OFLAG_IRCAD, 'I',
        OFLAG_TADMIN, 'T',
        OFLAG_COADMIN_, 'C',
        OFLAG_SCOADMIN_, 'S',
        OFLAG_ZLINE, 'z',
        OFLAG_WHOIS, 'W',
        OFLAG_HIDE, 'H',
        OFLAG_TKL, 't',
        OFLAG_GZL, 'Z',
        OFLAG_OVERRIDE, 'v',
        0, 0
};

CMD_FUNC(m_read_conf)
{
	//Using MSG_READCONF and TOK_READCONF

	aClient *cconfbuf;

	char *buffer;
	char *ptr;
	char *m = "*";
        int *i, flag;

	ConfigItem_oper *oper = NULL;
	ConfigItem_oper_from *from;
	anAuthStruct    *as = NULL;

	ConfigItem_ulines *ca;

	if (!parv[1])
	{
		sendto_snomask(SNO_CONF, " %s - CentralConf - Dropped Malformed Configuration", me.name);
		return 0;
	}

	if (!parv[2])
	{
		sendto_snomask(SNO_CONF, " %s - CentralConf - Missing Command", me.name);
		return 0;
	}

	if(stricmp(parv[1], me.name))
        {
                cconfbuf = find_server(parv[1], NULL);
                if (cconfbuf)
                        sendto_one(cconfbuf, ":%s READCONF %s :%s", sptr->name, cconfbuf->name, parv[2]);
                return 0;
        }

	if (IsServer(sptr))
	{
		switch (parv[2][0])
		{
			case 'O':
				sendto_snomask(SNO_CONF, " %s - CentralConf - Received O:Line from %s", me.name, parv[0]);
                                //O:User:Pass:Flags:Class:Snomask:Auth:Host:Host2
                                buffer = strtok_r(parv[2], ":", &ptr);
                                buffer = strtok_r(NULL, ":", &ptr);
				if (!buffer)
				{
				sendto_snomask(SNO_CONF, " %s - CentralConf - Failure Parsing Oper Name", me.name);
				return 0;
				}
				if (!(oper = Find_oper(buffer)))
				{
					oper =  MyMallocEx(sizeof(ConfigItem_oper));
					oper->name = strdup(buffer);
				}
				else
				{
					sendto_snomask(SNO_CONF, " %s - CentralConf - Duplicate O:Line Found", me.name);
					return 0;
				}
                                buffer = strtok_r(NULL, ":", &ptr);
				if (!buffer)
				{
				sendto_snomask(SNO_CONF, " %s - CentralConf - Failure Parsing Oper Password", me.name);
				ircfree(oper->name);
				MyFree(oper);
				return 0;
				}
				as = (anAuthStruct *) MyMalloc(sizeof(anAuthStruct));
				as->data = strdup(buffer);
				oper->auth = as;
                                buffer = strtok_r(NULL, ":", &ptr);
				if (!buffer)
				{
				sendto_snomask(SNO_CONF, " %s - CentralConf - Failure Parsing Oper Flags", me.name);
				ircfree(oper->name);
                                Auth_DeleteAuthStruct(oper->auth);
				MyFree(oper);
				return 0;
				}
                		for (m = (*buffer) ? buffer : m; *m; m++) {
                        		for (i = _OldOperFlags; (flag = *i); i += 2)
        	                        		if (*m == (char)(*(i + 1))) {
                	                        	oper->oflags |= flag;
                        		        	}
	                		}
				buffer = strtok_r(NULL, ":", &ptr);
				if (!buffer)
				{
				sendto_snomask(SNO_CONF, " %s - CentralConf - Failure Parsing Class", me.name);
				ircfree(oper->name);
                                Auth_DeleteAuthStruct(oper->auth);
				MyFree(oper);
				return 0;
				}
				oper->class = Find_class(buffer);
				if (!oper->class)
					oper->class = Find_class(REMOTECLASS);
				buffer = strtok_r(NULL, ":", &ptr);
				if (!buffer)
				{
				sendto_snomask(SNO_CONF, " %s - CentralConf - Failure Parsing Snomask", me.name);
				ircfree(oper->name);
                                Auth_DeleteAuthStruct(oper->auth);
				MyFree(oper);
				return 0;
				}
				ircstrdup(oper->snomask, buffer);
				buffer = strtok_r(NULL, ":", &ptr);
				if (!buffer)
				{
				sendto_snomask(SNO_CONF, " %s - CentralConf - Failure Parsing Authtype", me.name);
				ircfree(oper->name);
                                Auth_DeleteAuthStruct(oper->auth);
				ircfree(oper->snomask);
				MyFree(oper);
				return 0;
				}
				if (Auth_FindType(buffer) == -1)
				as->type = AUTHTYPE_PLAINTEXT;
				else
				as->type = Auth_FindType(buffer);
				buffer = strtok_r(NULL, ":", &ptr);
				if (!buffer)
				{
				sendto_snomask(SNO_CONF, " %s - CentralConf - Failure Parsing Oper Host", me.name);
				ircfree(oper->name);
                                Auth_DeleteAuthStruct(oper->auth);
                                ircfree(oper->snomask);
                                MyFree(oper);
				return 0;
				}
				while (buffer)
				{
					from = MyMallocEx(sizeof(ConfigItem_oper_from));
					ircstrdup(from->name, buffer);
					AddListItem(from, oper->from);
					buffer = strtok_r(NULL, ":", &ptr);
				}

                		AddListItem(oper, conf_oper);
				break;
			case 'U':
				//U:Server
				sendto_snomask(SNO_CONF, " %s - CentralConf - Received U:Line from %s", me.name, parv[0]);
				buffer = strtok_r(parv[2], ":", &ptr);
                                buffer = strtok_r(NULL, ":", &ptr);
				if (!buffer)
				{
				sendto_snomask(SNO_CONF, " %s - CentralConf - Failure Parsing U:Line", me.name);
				return 0;
				}
				ca = MyMallocEx(sizeof(ConfigItem_ulines));
				ircstrdup(ca->servername, buffer);
				AddListItem(ca, conf_ulines);
				break;
			default:
				sendto_snomask(SNO_CONF, " %s - CentralConf - Unknown [Case: %c]", me.name, parv[2][0]);
				break;
		}
	}
	return 0;
}
#endif

CMD_FUNC(m_ircops)
{
    aClient *acptr;
    int totalops = 0;
    char *stat, buf[200];

    /* if (!IsOper(cptr)) { */
    sendto_one(cptr, ":%s 009 %s :  %s On-Line Operator List  ",
        me.name, cptr->name, ircnet005);
    sendto_one(cptr, ":%s 009 %s :                                                                       ",
        me.name, cptr->name);
    sendto_one(cptr, ":%s 009 %s :\2Nick\2                           Status                      \02Server\02", me.name, cptr->name);
    sendto_one(cptr, ":%s 009 %s :",
                me.name, cptr->name);

    for (acptr = client; acptr; acptr = acptr->next)
    {
        if (IsPerson(acptr) && IsAnOper(acptr) && !IsULine(acptr) && !IsHideOper(acptr))
        {
            buf[0] = '\0';
            sendto_one(cptr, ":%s 009 %s :%-30s  %-9s %-9s  %s",me.name, cptr->name,
                    acptr->name ? acptr->name : "<Desconhecido>",
		    IsIRCAdmin(acptr) ? "IRC Administrator" : IsNetAdmin(acptr) ? "Network Administrator" : IsCoAdmin(acptr) ? "Co-Administrator" : IsTechAdmin(acptr) ? "Technical Administrator" : IsSAdmin(acptr) ? "Services Administrator" : IsAdmin(acptr) ? "Server Administrator" : IsServerCoAdmin(acptr) ? "Server Co-Administrator" : IsOper(acptr) ? "Global Operator" : "Local Operator",
		    acptr->user->away ? "(Away)" : "",
		    acptr->user->server);

	    totalops++;
	}
    }

    if (totalops == 0)
        sendto_one(cptr, ":%s 009 %s :No On-Line Opers ",me.name, cptr->name);
    sendto_one(cptr, ":%s 009 %s :",
        me.name, cptr->name);
    sendto_one(cptr, ":%s 009 %s :Total:  %d  Operator(s) connected.", me.name, cptr->name, totalops);
    sendto_one(cptr, ":%s 009 %s :",
        me.name, cptr->name);
}

/*
** m_relay
**      parv[0] = sender
**      parv[1] = server pattern to relay to
**      parv[2] = string to relay
*/

CMD_FUNC(m_relay)
{
        Link *lp;

        if (!(IsNetAdmin(sptr) || IsCoAdmin(sptr) || IsServer(cptr) || IsULine(sptr))) {
                sendto_one(sptr, err_str(ERR_NOPRIVILEGES), me.name, parv[0]);
                return -1;
        }

        if (parc < 3 || *parv[1] == '\0')
        {
                sendto_one(sptr, err_str(ERR_NEEDMOREPARAMS), me.name, parv[0], "RELAY");
                return -1;
        }

        if (!IsServer(cptr)) {
                sendto_serv_butone(&me,
                    ":%s GLOBOPS :%s is RELAYing [%s] to %s",
                    me.name, sptr->name, parv[2], parv[1]);
                sendto_realops
                    ("%s is RELAYing [%s] to %s",
                    parv[0], parv[2], parv[1]);
        }

        if (hunt_server(cptr, sptr, ":%s RELAY %s :%s", 1, parc, parv) == HUNTED_ISME) {
                parse(sptr, parv[2], (parv[2] + strlen(parv[2])));
        }

        return 0;
}

/*
** m_grelay
**      parv[0] = sender
**      parv[1] = string to relay
*/

CMD_FUNC(m_grelay)
{
        Link *lp;

        if (!(IsNetAdmin(sptr) || IsCoAdmin(sptr) || IsULine(sptr))) {
                sendto_one(sptr, err_str(ERR_NOPRIVILEGES), me.name, parv[0]);
                return -1;
        }

        if (parc < 2 || *parv[1] == '\0')
        {
                sendto_one(sptr, err_str(ERR_NEEDMOREPARAMS), me.name, parv[0], "GRELAY");
                return -1;
        }

        sendto_serv_butone(&me,
            ":%s GLOBOPS :%s is GRELAYing [%s]",
            me.name, sptr->name, parv[1]);
        sendto_realops
            ("%s is GRELAYing [%s]",
            parv[0], parv[1]);


        parv[2] = parv[1];
        parc++;

        for (lp = Servers; lp; lp = lp->next) {
                parv[1] = lp->value.cptr->name;
                if (hunt_server(cptr, sptr, ":%s RELAY %s :%s", 1, parc, parv) == HUNTED_ISME) {
                        parse(sptr, parv[2], (parv[2] + strlen(parv[2])));
                        break;
                }
        }

        return 0;
}

/*
** m_glag
**      parv[0] = sender
*/

CMD_FUNC(m_glag)
{
        Link *lp;

        if (!IsOper(sptr)) {
                sendto_one(sptr, err_str(ERR_NOPRIVILEGES), me.name, parv[0]);
                return -1;
        }

        for (lp = Servers; lp; lp = lp->next) {
                parv[1] = lp->value.cptr->name;
                if (hunt_server(cptr, sptr, ":%s RELAY %s :LAG", 1, 2, parv) == HUNTED_ISME) {
                        parv[2] = strdup("LAG");
                        parse(sptr, parv[2], (parv[2] + 3));
                        free(parv[2]);
                        break;
                }
        }

        return 0;
}
